#!/usr/bin/python


from kernel import *
from parser import parseprogram, parseterm

"""
 Goals:
 	- Given a program and a specification
		- Computationally check the verification for some inputs
		- Formal proof
	- Given a specification
		- Find a program that satisfy it
		- Generate instances to test the specification
		- Make a tester for such specification

 Ideas:
 	- Language to define datatypes
		- And its destructors

 Pages with programming problems:
	http://programmers.stackexchange.com/questions/756/where-can-i-find-programming-puzzles-and-challenges
	http://www.spoj.com/problems/classical/sort=11
	http://codingbat.com/
	https://en.wikipedia.org/wiki/List_of_algorithms
"""




def example1a():
	"""
	@output := @input

	postcondition: @input=@output
	"""
	print "=== Example 1 ==="
	a=Assignment(Ass_Variable("@output"),E_Variable("@input"))
	p=Program([a])
	print p.execute(DS_String("Hola"))
	condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	p.randomcheck(condition)

def example1b():
	"""
	@output := @input

	postcondition: @input=@output
	"""
	print "=== Example 1 ==="
	p=parseprogram("@output := @input")
	print p
	print p.execute(DS_String("Hola"))
	#condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	condition=parseterm("@input=@output")
	print condition
	p.randomcheck(condition)

def example2a():
	"""
	i := 0
	while @input[i] not null:
		@output[i] := @input[i] + i
	
	postcondition: isarray(@input) -> forall i:nat, i<length(@input) -> @output[i]=@input[i]+i
	"""
	print "=== Example 2 ==="
	m=DS_Array()
	#m.value.append(DS_Integer(42))
	i0=Assignment(Ass_Variable("i"),E_Constant(DS_Integer(0)))
	a_init=Assignment(Ass_Variable("@output"),E_Constant(m))
	#a_copy=Assignment(Ass_Access("@output",E_Variable("i")),E_Access("@input",E_Variable("i")))
	a_copy=Assignment(Ass_Access("@output",E_Variable("i")),
		E_Addition( [E_Access("@input",E_Variable("i")) , E_Variable("i")] ) )
	#w=While(E_Constant(DS_Bool(True)),[a_copy])
	i_inc=Assignment(Ass_Variable("i"),E_Addition([E_Variable("i"),E_Constant(DS_Integer(1))]))
	w=While(E_Not(E_IsNull(E_Access("@input",E_Variable("i")))),[a_copy,i_inc])
	p=Program([i0, a_init, w])
	l=DS_Array()
	for i in range(5):
		l.value.append(DS_Integer(i+3))
	print p.execute(l)
	#condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	#postcondition: isarray(@input) -> forall i:nat, @output[i]=@input[i]+i
	#condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Array),Product("i",DS_Integer,E_Equal(E_Access("@output",E_Variable("i")),E_Addition([E_Access("@input",E_Variable("i")),E_Variable("i")]))))
	#postcondition: isarray(@input) -> forall i:nat, i<length(@input) -> @output[i]=@input[i]+i
	condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Array),Product("i",DS_Integer,
		Product("y",E_LT(E_Variable("i"),E_Length(E_Variable("@input"))),
		E_Equal(E_Access("@output",E_Variable("i")),E_Addition([E_Access("@input",E_Variable("i")),E_Variable("i")]))
		)))
	p.randomcheck(condition)

def example2b():
	print "=== Example 2 ==="
	p=parseprogram("""
i := 0
@output := []
while not (isnull @input[i])
{
	@output[i] := @input[i] + i
	i := i + 1
}
""")
	print p
	l=DS_Array()
	for i in range(5):
		l.value.append(DS_Integer(i+3))
	print "@input",l
	print "@output",p.execute(l)
	#condition=parseterm("isdatatype @input DS_Array -> forall i:DS_Integer, i<(length @input) -> @output[i]=@input[i]+i")
	#condition=parseterm("isdatatype @input DS_Array -> ((length @input)=(length @output)) and forall i:DS_Integer, (i<(length @input)) -> @output[i]=@input[i]+i")
	#condition=parseterm("isdatatype @input DS_Array -> length @input=length @output and forall i:DS_Integer, i<(length @input) -> @output[i]=@input[i]+i")
	condition=parseterm("isdatatype @input DS_Array -> length @input=length @output and forall i:DS_Integer, 0<=i and i<(length @input) -> @output[i]=@input[i]+i")
	print condition
	p.randomcheck(condition)

def example2gen():
	print "=== Example 2 gen ==="
	#condition=parseterm("isdatatype @input DS_Array -> forall i:DS_Integer, i<(length @input) -> @output[i]=@input[i]+i")
	#condition=parseterm("isdatatype @input DS_Array -> length @input=length @output and forall i:DS_Integer, i<(length @input) -> @output[i]=@input[i]+i")
	condition=parseterm("isdatatype @input DS_Array -> forall i:DS_Integer, 0<=i and i<(length @input) -> @output[i]=@input[i]+i")
	print "Looking programs for condition \"%s\"" % condition
	p=generateprogram(condition)
	print p
	print "Testing condition"
	p.randomcheck(condition)

def example3a():
	"""
	i:=0
	@output:=[]
	while i<@input
		@output[i]:=i
		i:=i+1

	postcondition: isinteger(@input) -> forall n:nat, n<@input -> @output[n]=n
	"""
	print "=== Example 3 ==="
	i_init=Assignment(Ass_Variable("i"),E_Constant(DS_Integer(0)))
	a_init=Assignment(Ass_Variable("@output"),E_Constant(DS_Array()))
	a_set=Assignment(Ass_Access("@output",E_Variable("i")),E_Variable("i"))
	i_inc=Assignment(Ass_Variable("i"),E_Addition([E_Variable("i"),E_Constant(DS_Integer(1))]))
	w=While(E_LT(E_Variable("i"),E_Variable("@input")),[a_set, i_inc])
	p=Program([i_init,a_init,w])
	print p.execute(DS_Integer(5))
	print p.execute(DS_null)
	condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Integer),Product("n",DS_Integer,Product("y",E_LT(E_Variable("n"),E_Variable("@input")),E_Equal(E_Access("@output",E_Variable("n")),E_Variable("n")))))
	p.randomcheck(condition)

def example3b():
	print "=== Example 3 ==="
	p=parseprogram("""
i := 0
@output := []
while i < @input
{
	@output[i] := i
	i := i + 1
}
""")
	print p
	print "@input=5, @output=",p.execute(DS_Integer(5))
	print "@input=null, @output=",p.execute(DS_null)
	#condition=parseterm("isdatatype @input DS_Integer -> forall n:DS_Integer, n<@input -> @output[n]=n")
	#condition=parseterm("isdatatype @input DS_Integer -> 0<=@input -> length @output=@input and forall n:DS_Integer, n<@input -> @output[n]=n")
	condition=parseterm("isdatatype @input DS_Integer -> 0<=@input -> length @output=@input and forall n:DS_Integer, 0<=n and n<@input -> @output[n]=n")
	print condition
	p.randomcheck(condition)

def example3gen():
	print "=== Example 3 gen ==="
	condition=parseterm("isdatatype @input DS_Integer -> 0<=@input -> length @output=@input and forall n:DS_Integer, 0<=n and n<@input -> @output[n]=n")
	print "Looking programs for condition \"%s\"" % condition
	p=generateprogram(condition)
	print p
	print "Testing condition"
	p.randomcheck(condition)


def example4():
	"""
		Reverse the input
	"""
	print "=== Example 4 ==="
	p=parseprogram("""
i := 0
j := length @input - 1
@output := []
while i < length @input
{
	@output[i] := @input[j]
	i := i + 1
	j := j - 1
}
""")
	print p
	l=[1,2,6,7,9]
	print "@input=",l,"@output=",p.execute(DS_Array([DS_Integer(x) for x in l]))
	print "@input=null, @output=",p.execute(DS_null)
	#condition=parseterm("isdatatype @input DS_Array -> forall n:DS_Integer, n<length @input -> @output[n]=@input[length @input -n -1]")
	condition=parseterm("isdatatype @input DS_Array -> length @output=length @input and forall n:DS_Integer, n<length @input -> @output[n]=@input[length @input -n -1]")
	print condition
	p.randomcheck(condition)

def example4gen():
	print "=== Example 4 gen ==="
	#condition=parseterm("isdatatype @input DS_Array -> length @output=length @input and forall n:DS_Integer, n<length @input -> @output[n]=@input[length @input -n -1]")#inferring that body implies 0<=n is hard
	condition=parseterm("isdatatype @input DS_Array -> length @output=length @input and forall n:DS_Integer, n<length @input and 0<=n -> @output[n]=@input[length @input -n -1]")
	print "Looking programs for condition \"%s\"" % condition
	p=generateprogram(condition)
	print p
	print "Testing condition"
	p.randomcheck(condition)


def example5gen():
	"""
	Duplicate the input
	"""
	print "=== Example 5 gen ==="
	condition=parseterm("isdatatype @input DS_Array -> length @output=length @input + length @input and forall n:DS_Integer, n<length @input and 0<=n -> @output[n]=@input[n] and @output[n+length @input]=@input[n]")
	print "Looking programs for condition \"%s\"" % condition
	p=generateprogram(condition)
	print p
	print "Testing condition"
	p.randomcheck(condition)


example_conditions=[
	None,
	#1
	parseterm("@input=@output"),
	#2
	parseterm("isdatatype @input DS_Array -> forall i:DS_Integer, 0<=i and i<(length @input) -> @output[i]=@input[i]+i"),
	#3
	parseterm("isdatatype @input DS_Integer -> 0<=@input -> length @output=@input and forall n:DS_Integer, 0<=n and n<@input -> @output[n]=n"),
	#4
	parseterm("isdatatype @input DS_Array -> length @output=length @input and forall n:DS_Integer, n<length @input and 0<=n -> @output[n]=@input[length @input -n -1]"),
	#5
	parseterm("isdatatype @input DS_Array -> length @output=length @input + length @input and forall n:DS_Integer, n<length @input and 0<=n -> @output[n]=@input[n] and @output[n+length @input]=@input[n]"),
	#6
	#This  have some complications without using polyhedral calculations or a rational upper bound
	parseterm("isdatatype @input DS_Array -> forall i:DS_Integer, 0<=i and (i+i<length @input or i<length @output) -> @output[i]=@input[i+i]"),
	#7
	#return a*[b]
	#This is not very well verified randomly...
	parseterm('isdatatype @input DS_Dictionary -> isdatatype @input["a"] DS_Integer -> length @output=@input["a"] and forall i:DS_Integer, 0<=i and i<@input["a"] -> @output[i]=@input["b"]'),
]

example_code=[
	None,
	#1
	parseprogram("@output := @input"),
	#2
	parseprogram("""
i := 0
@output := []
while not (isnull @input[i])
{
	@output[i] := @input[i] + i
	i := i + 1
}
"""),
	#3
	parseprogram("""
i := 0
@output := []
while i < @input
{
	@output[i] := i
	i := i + 1
}
"""),
	#4
	parseprogram("""
i := 0
j := length @input - 1
@output := []
while i < length @input
{
	@output[i] := @input[j]
	i := i + 1
	j := j - 1
}
"""),
	#5
	None,
	#6
#	parseprogram("""
#i:=0
#j:=0
#@output:=[]
#while i<length @input
#{
#	@output[j]:=@input[i]
#	i:=i+2
#	j:=j+1
#}
#""")
	parseprogram("""
i:=0
@output:=[]
while i+i<length @input
{
	@output[i]:=@input[i+i]
	i:=i+1
}
"""),
	#7
	parseprogram("@output:=@input"),#FIXME: This should fail...
]


def example_test(index=None,program=None,condition=None):
	if index is None:
		print "=== Example testing ==="
		if program is None: raise Exception()
		if condition is None: raise Exception()
	else:
		print "=== Example testing %d ===" % index
		if program is not None: raise Exception()
		if condition is not None: raise Exception()
		program=example_code[index]
		condition=example_conditions[index]
	print "PROGRAM:"
	print program
	print "CONDITION:"
	print condition
	print "TEST RESULTS:"
	program.randomcheck(condition)

def example_gen(index=None,condition=None):
	if index is None:
		print "=== Example generation ==="
		if condition is None: raise Exception()
	else:
		print "=== Example generation %d ===" % index
		if condition is not None: raise Exception()
		condition=example_conditions[index]
	#print "Looking programs for condition \"%s\"" % condition
	print "CONDITION:"
	print condition
	p=generateprogram(condition)
	print "GENERATED PROGRAM:"
	print p
	#print "Testing condition"
	print "TEST RESULTS:"
	p.randomcheck(condition)

if __name__=="__main__":
	example_test(7)
	#example_gen(7)


"""
- 6: Get even values of a sequence		http://codingbat.com/prob/p113152
	return "".join((str[i] for i in range(len(str)) if i%2==0))
	forall i:nat, 2*i<len(@input) or i<len(@output) -> @output[i]=@input[2*i]
- Count instances		http://codingbat.com/prob/p166170
	return len([0 for x in nums if x==9])
- Count substrings		http://codingbat.com/prob/p145834	http://codingbat.com/prob/p167246
- Given two vectors say if their sum is the zero vector
	@output <-> (len(@input["a"])=len(@input["b"])) and forall i:nat, @input["a"][i]+@input["b"][i]=0
- Given integers a,b return True if 6 is in a,b,a+b,a-b		http://codingbat.com/prob/p100958
- Given integer n, return True if it is distance 2 from some multiple of 10.
	That is "return ((n+2)%10)<=4"		http://codingbat.com/prob/p165321
- Reduce array until mark		http://codingbat.com/prob/p107863
- Reduce with filter			http://codingbat.com/prob/p100347
- Average of array discarding extremes		http://codingbat.com/prob/p126968
- Reduce discarding lexer tokens		http://codingbat.com/prob/p108886


EXAMENES CARMEN:
	- Given string zero terminated and word aligned and pointer for result also word aligned.
		Break the string (exclude the terminating 0) into blocks of 4 chars, the last block may be uncomplete, consider it filled with zeros
		For each block [a,b,c,d] compute f(a,b,c,d), then write in the result the word [a,b,c,d] followed by the word f([a,b,c,d])
		write a terminating 0 (char)
		Reading outside of the input string gives undefined behavious (might be a segfault).
		let s_1,...,s_n be such that @input=s_1+...+s_n, length s_i=4 for i<n and 1<=length s_n<=4
		let c:string->string be such that length c(x)=4 and c(x)=0+...+0+x for some amount of 0s
		The @output must be s_1+f(s_1)+s_2+f(s_2)+...+s_{n-1}+f(s_{n-1})+c(s_n)+f(c(s_n))+"\0"
"""

