#!/usr/bin/python

import random
import string

class DataStructure:
	def get(self,position):
		return DS_null
	def copy(self):
		#return self
		return self.__class__(self.value)
	@staticmethod
	def randomclass():
		return random.choice(DataStructure.subclasses)
	@classmethod
	def random(cls):
		if cls==DataStructure:
			return cls.randomclass().random()
		else:
			raise NotImplementedError("random() not implemented for class %s" % cls)
class DS_NULL(DataStructure):
	@classmethod
	def random(cls):
		return DS_null
DS_null=DS_NULL()
class DS_Bool(DataStructure):
	def __init__(self,value):
		if not isinstance(value,bool): raise ValueError
		self.value=value
	@classmethod
	def random(cls):
		return DS_Bool(random.choice([True,False]))
DS_true=DS_Bool(True)
DS_false=DS_Bool(False)
class DS_Error(DataStructure):
	@classmethod
	def random(cls):
		return DS_error
DS_error=DS_Error()
class DS_Integer(DataStructure):
	def __init__(self,value):
		if not isinstance(value,int): raise ValueError
		self.value=value
	def __str__(self):
		return str(self.value)
	@classmethod
	def random(cls):
		return DS_Integer(random.randrange(-1000,1000))#FIXME
class DS_String(DataStructure):
	def __init__(self,value):
		if not isinstance(value,str): raise ValueError
		self.value=value
	def __str__(self):
		return self.value
	@classmethod
	def random(cls):
		#http://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits-in-python
		s=''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(100))#FIXME
		return DS_String(s)
class DS_Array(DataStructure):
	def __init__(self,value=None):
		if value is None:
			self.value=[]
		else:
			self.value=value[:]
	def __str__(self):
		return "["+", ".join(str(x) for x in self.value)+"]"
	def writeat(self,position,value):
		while position>=len(self.value):
			self.value.append(DS_null)
		self.value[position]=value
	def get(self,position):
		if position<len(self.value):
			return self.value[position]
		else:
			return DS_null
	def copy(self):
		return self.__class__(self.value[:])
	@classmethod
	def random(cls):
		n=random.randrange(0,1000)#FIXME
		l=[DS_Integer.random() for t in range(n)]
		return DS_Array(l)
DataStructure.subclasses=[DS_NULL,DS_Bool,DS_Integer,DS_String,DS_Array]

class Assignable: pass #a.k.a. Left value
class Ass_Variable(Assignable):
	def __init__(self,variable):
		self.variable=variable
	def write(self,variables,value):
		variables[self.variable]=value
class Ass_Access(Assignable):
	def __init__(self,variable,position):
		self.variable=variable #string
		if not isinstance(position,Expression): raise ValueError()
		self.position=position
	def write(self,variables,value):
		pos=self.position.evaluate(variables)
		if not isinstance(pos,DS_Integer):
			raise Exception("%s:%s cannot be used to index arrays" % (pos,pos.__class__))
		#variables[self.variable].value[pos.value]=value
		variables[self.variable].writeat(pos.value,value)

class Term: pass

class Expression(Term): pass
class E_Constant(Expression):
	def __init__(self,constant):
		if not isinstance(constant,DataStructure): raise ValueError()
		self.constant=constant
	def evaluate(self,variables):
		return self.constant
class E_Variable(Expression):
	def __init__(self,variable):
		if not isinstance(variable,str): raise ValueError()
		self.variable=variable
	def evaluate(self,variables):
		if self.variable not in variables:
			raise Exception("There is not variable '%s'" % self.variable)
		return variables[self.variable]
	def __str__(self):
		return self.variable
class E_Access(Expression):
	def __init__(self,variable,position):
		self.variable=variable #string
		if not isinstance(position,Expression): raise ValueError()
		self.position=position
	def evaluate(self,variables):
		if self.variable not in variables:
			raise Exception("There is not variable '%s'" % self.variable)
		pos=self.position.evaluate(variables)
		if not isinstance(pos,DS_Integer):
			raise Exception("%s:%s cannot be used to index arrays" % (pos,pos.__class__))
		return variables[self.variable].get(pos.value)
	def __str__(self):
		return self.variable+"["+str(self.position)+"]"
class E_Addition(Expression):
	def __init__(self,le):
		self.le=le
	def evaluate(self,variables):
		#raise Exception("Implement")
		a=0
		for e in self.le:
			data=e.evaluate(variables)
			if not isinstance(data,DS_Integer): raise Exception("%s cannot be added" % data)
			a+=data.value
		return DS_Integer(a)
	def __str__(self):
		return "+".join( "["+str(e)+"]" for e in self.le )
class E_IsNull(Expression):
	def __init__(self,expression):
		self.expression=expression
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		return DS_Bool(isinstance(data,DS_NULL))
class E_IsDataType(Expression):
	def __init__(self,expression,datatype):
		self.expression=expression
		self.datatype=datatype
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		return DS_Bool(data.__class__==self.datatype)
class E_Not(Expression):
	def __init__(self,expression):
		self.expression=expression
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		if not isinstance(data,DS_Bool): raise Exception()
		return DS_Bool(not data.value)
class E_Equal(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		data1=self.expression1.evaluate(variables)
		data2=self.expression2.evaluate(variables)
		#print "%s(%s)=%s(%s) : %s" % (data1,data1.__class__,data2,data2.__class__,data1==data2)
		if data1.__class__ != data2.__class__: return DS_false
		return DS_Bool(data1.value==data2.value)
	def __str__(self):
		s1=str(self.expression1)
		s2=str(self.expression2)
		return "("+s1+")=("+s2+")"
class E_LT(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		data1=self.expression1.evaluate(variables)
		data2=self.expression2.evaluate(variables)
		if not isinstance(data1,DS_Integer): return DS_null
		if not isinstance(data2,DS_Integer): return DS_null
		#print "%s(%s)=%s(%s) : %s" % (data1,data1.__class__,data2,data2.__class__,data1==data2)
		#if data1.__class__ != data2.__class__: return DS_false
		return DS_Bool(data1.value<data2.value)
class E_Length(Expression):
	def __init__(self,expression):
		self.expression=expression
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		if not isinstance(data,DS_Array): raise Exception()
		return DS_Integer(len(data.value))

class BoundTerm(Term):
	def __init__(self,var,vartype,body):
		if self.__class__==BoundTerm:
			raise ValueError("You cannot instantiate a BoundTerm")
		self.variable=var
		self.vartype=vartype
		self.body=body
		#bfv=set(body.getfreevars())
		#bfv.discard(var.name)
		#self.freevars=bfv|vartype.getfreevars()

class Product(BoundTerm):
	"""
	forall var:vartype, body
	"""
	def evaluate(self,variables):
		head=self.vartype
		headdata=None
		if isinstance(head,Expression):
			headdata=self.vartype.evaluate(variables)
		#print "vartype=%s" % data
		if isinstance(headdata,DS_Bool):
			if not headdata.value: return DS_true #False -> anything
			return self.body.evaluate(variables)
		elif head==DS_Integer:
			#FIXME: induction
			for i in range(100):
				d=dict(variables)
				#print d["@input"].value[0],d["@output"].value[0]
				variables[self.variable]=DS_Integer(i)
				#print self.body
				r=self.body.evaluate(variables)
				#print r.value
				if not isinstance(r,DS_Bool): raise Exception()
				if not r.value: return DS_false
			return DS_true
		else:
			raise Exception("Cannot evaluate such product (%s:%s)" % (data,data.__class__))

class Abstraction(BoundTerm): pass
class Application(Term): pass

class Statement: pass
class Assignment(Statement):
	def __init__(self,ass,expression):
		# ass := expression
		if not isinstance(ass,Assignable): raise ValueError("ass=%s:%s should be an Assignable" % (ass,ass.__class__))
		self.ass=ass
		self.expression=expression
	def execute(self,variables):
		#variables[self.variable]=self.expression.evaluate(variables)
		self.ass.write(variables,self.expression.evaluate(variables).copy())
class While(Statement):
	def __init__(self,condition,body):
		if not isinstance(condition,Expression): raise ValueError()
		self.condition=condition
		self.body=body #list of Statements
	def execute(self,variables):
		while True:
			e=self.condition.evaluate(variables)
			if not isinstance(e,DS_Bool): return DS_error
			if not e.value: return
			for s in self.body:
				s.execute(variables)


class Program:
	def __init__(self,body):
		self.body=body #list of Statement
	def execute(self,input):
		variables={"@input":input}
		for s in self.body:
			r=s.execute(variables)
			if isinstance(r,DS_Error): return r
		return variables["@output"]
	def randomcheck(self,condition,n=100):
		good=0
		for i in range(n):
			input=DataStructure.random()
			#print "INPUT:",input
			output=self.execute(input)
			variables={"@input":input,"@output":output}
			#print "OUTPUT:",output,id(output)
			e=condition.evaluate(variables)
			if isinstance(e,DS_Bool) and e.value:
				good+=1
				#print "isgood"
		print "good = %d/%d (%f)" % (good,n,float(good)/n)


def example1():
	"""
	@output := @input

	postcondition: @input=@output
	"""
	print "=== Example 1 ==="
	a=Assignment(Ass_Variable("@output"),E_Variable("@input"))
	p=Program([a])
	print p.execute(DS_String("Hola"))
	condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	p.randomcheck(condition)

def example2():
	"""
	i := 0
	while @input[i] not null:
		@output[i] := @input[i] + i
	
	postcondition: isarray(@input) -> forall i:nat, i<length(@input) -> @output[i]=@input[i]+i
	"""
	print "=== Example 2 ==="
	m=DS_Array()
	#m.value.append(DS_Integer(42))
	i0=Assignment(Ass_Variable("i"),E_Constant(DS_Integer(0)))
	a_init=Assignment(Ass_Variable("@output"),E_Constant(m))
	#a_copy=Assignment(Ass_Access("@output",E_Variable("i")),E_Access("@input",E_Variable("i")))
	a_copy=Assignment(Ass_Access("@output",E_Variable("i")),
		E_Addition( [E_Access("@input",E_Variable("i")) , E_Variable("i")] ) )
	#w=While(E_Constant(DS_Bool(True)),[a_copy])
	i_inc=Assignment(Ass_Variable("i"),E_Addition([E_Variable("i"),E_Constant(DS_Integer(1))]))
	w=While(E_Not(E_IsNull(E_Access("@input",E_Variable("i")))),[a_copy,i_inc])
	p=Program([i0, a_init, w])
	l=DS_Array()
	for i in range(5):
		l.value.append(DS_Integer(i+3))
	print p.execute(l)
	#condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	#postcondition: isarray(@input) -> forall i:nat, @output[i]=@input[i]+i
	#condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Array),Product("i",DS_Integer,E_Equal(E_Access("@output",E_Variable("i")),E_Addition([E_Access("@input",E_Variable("i")),E_Variable("i")]))))
	#postcondition: isarray(@input) -> forall i:nat, i<length(@input) -> @output[i]=@input[i]+i
	condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Array),Product("i",DS_Integer,
		Product("y",E_LT(E_Variable("i"),E_Length(E_Variable("@input"))),
		E_Equal(E_Access("@output",E_Variable("i")),E_Addition([E_Access("@input",E_Variable("i")),E_Variable("i")]))
		)))
	p.randomcheck(condition)

def example3():
	"""
	i:=0
	@output:=[]
	while i<@input
		@output[i]:=i
		i:=i+1

	postcondition: isinteger(@input) -> forall n:nat, n<@input -> @output[n]=n
	"""
	print "=== Example 3 ==="
	i_init=Assignment(Ass_Variable("i"),E_Constant(DS_Integer(0)))
	a_init=Assignment(Ass_Variable("@output"),E_Constant(DS_Array()))
	a_set=Assignment(Ass_Access("@output",E_Variable("i")),E_Variable("i"))
	i_inc=Assignment(Ass_Variable("i"),E_Addition([E_Variable("i"),E_Constant(DS_Integer(1))]))
	w=While(E_LT(E_Variable("i"),E_Variable("@input")),[a_set, i_inc])
	p=Program([i_init,a_init,w])
	print p.execute(DS_Integer(5))
	print p.execute(DS_null)
	condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Integer),Product("n",DS_Integer,Product("y",E_LT(E_Variable("n"),E_Variable("@input")),E_Equal(E_Access("@output",E_Variable("n")),E_Variable("n")))))
	p.randomcheck(condition)

if __name__=="__main__":
	example3()

