#!/usr/bin/python


from kernel import *

from parser import parseprogram, parseterm

"""
 Goals:
 	- Given a program and a specification
		- Computationally check the verification for some inputs
		- Formal proof
	- Given a specification
		- Find a program that satisfy it
		- Generate instances to test the specification
		- Make a tester for such specification

http://programmers.stackexchange.com/questions/756/where-can-i-find-programming-puzzles-and-challenges
http://www.spoj.com/problems/classical/sort=11
http://codingbat.com/
"""




def example1a():
	"""
	@output := @input

	postcondition: @input=@output
	"""
	print "=== Example 1 ==="
	a=Assignment(Ass_Variable("@output"),E_Variable("@input"))
	p=Program([a])
	print p.execute(DS_String("Hola"))
	condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	p.randomcheck(condition)

def example1b():
	"""
	@output := @input

	postcondition: @input=@output
	"""
	print "=== Example 1 ==="
	p=parseprogram("@output := @input")
	print p
	print p.execute(DS_String("Hola"))
	#condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	condition=parseterm("@input=@output")
	print condition
	p.randomcheck(condition)

def example2a():
	"""
	i := 0
	while @input[i] not null:
		@output[i] := @input[i] + i
	
	postcondition: isarray(@input) -> forall i:nat, i<length(@input) -> @output[i]=@input[i]+i
	"""
	print "=== Example 2 ==="
	m=DS_Array()
	#m.value.append(DS_Integer(42))
	i0=Assignment(Ass_Variable("i"),E_Constant(DS_Integer(0)))
	a_init=Assignment(Ass_Variable("@output"),E_Constant(m))
	#a_copy=Assignment(Ass_Access("@output",E_Variable("i")),E_Access("@input",E_Variable("i")))
	a_copy=Assignment(Ass_Access("@output",E_Variable("i")),
		E_Addition( [E_Access("@input",E_Variable("i")) , E_Variable("i")] ) )
	#w=While(E_Constant(DS_Bool(True)),[a_copy])
	i_inc=Assignment(Ass_Variable("i"),E_Addition([E_Variable("i"),E_Constant(DS_Integer(1))]))
	w=While(E_Not(E_IsNull(E_Access("@input",E_Variable("i")))),[a_copy,i_inc])
	p=Program([i0, a_init, w])
	l=DS_Array()
	for i in range(5):
		l.value.append(DS_Integer(i+3))
	print p.execute(l)
	#condition=E_Equal(E_Variable("@input"),E_Variable("@output"))
	#postcondition: isarray(@input) -> forall i:nat, @output[i]=@input[i]+i
	#condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Array),Product("i",DS_Integer,E_Equal(E_Access("@output",E_Variable("i")),E_Addition([E_Access("@input",E_Variable("i")),E_Variable("i")]))))
	#postcondition: isarray(@input) -> forall i:nat, i<length(@input) -> @output[i]=@input[i]+i
	condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Array),Product("i",DS_Integer,
		Product("y",E_LT(E_Variable("i"),E_Length(E_Variable("@input"))),
		E_Equal(E_Access("@output",E_Variable("i")),E_Addition([E_Access("@input",E_Variable("i")),E_Variable("i")]))
		)))
	p.randomcheck(condition)

def example2b():
	print "=== Example 2 ==="
	p=parseprogram("""
i := 0
@output := []
while not (isnull @input[i])
{
	@output[i] := @input[i] + i
	i := i + 1
}
""")
	print p
	l=DS_Array()
	for i in range(5):
		l.value.append(DS_Integer(i+3))
	print "@input",l
	print "@output",p.execute(l)
	#condition=parseterm("isdatatype @input DS_Array -> forall i:DS_Integer, i<(length @input) -> @output[i]=@input[i]+i")
	#condition=parseterm("isdatatype @input DS_Array -> ((length @input)=(length @output)) and forall i:DS_Integer, (i<(length @input)) -> @output[i]=@input[i]+i")
	condition=parseterm("isdatatype @input DS_Array -> ((length @input)=(length @output)) and forall i:DS_Integer, i<(length @input) -> @output[i]=@input[i]+i")
	print condition
	p.randomcheck(condition)

def example3a():
	"""
	i:=0
	@output:=[]
	while i<@input
		@output[i]:=i
		i:=i+1

	postcondition: isinteger(@input) -> forall n:nat, n<@input -> @output[n]=n
	"""
	print "=== Example 3 ==="
	i_init=Assignment(Ass_Variable("i"),E_Constant(DS_Integer(0)))
	a_init=Assignment(Ass_Variable("@output"),E_Constant(DS_Array()))
	a_set=Assignment(Ass_Access("@output",E_Variable("i")),E_Variable("i"))
	i_inc=Assignment(Ass_Variable("i"),E_Addition([E_Variable("i"),E_Constant(DS_Integer(1))]))
	w=While(E_LT(E_Variable("i"),E_Variable("@input")),[a_set, i_inc])
	p=Program([i_init,a_init,w])
	print p.execute(DS_Integer(5))
	print p.execute(DS_null)
	condition=Product("x",E_IsDataType(E_Variable("@input"),DS_Integer),Product("n",DS_Integer,Product("y",E_LT(E_Variable("n"),E_Variable("@input")),E_Equal(E_Access("@output",E_Variable("n")),E_Variable("n")))))
	p.randomcheck(condition)

def example3b():
	print "=== Example 3 ==="
	p=parseprogram("""
i := 0
@output := []
while i < @input
{
	@output[i] := i
	i := i + 1
}
""")
	print p
	print "@input=5, @output=",p.execute(DS_Integer(5))
	print "@input=null, @output=",p.execute(DS_null)
	condition=parseterm("isdatatype @input DS_Integer -> forall n:DS_Integer, n<@input -> @output[n]=n")
	print condition
	p.randomcheck(condition)

def example4():
	"""
		Reverse the input
	"""
	print "=== Example 4 ==="
	p=parseprogram("""
i := 0
j := length @input - 1
@output := []
while i < length @input
{
	@output[i] := @input[j]
	i := i + 1
	j := j - 1
}
""")
	print p
	l=[1,2,6,7,9]
	print "@input=",l,"@output=",p.execute(DS_Array([DS_Integer(x) for x in l]))
	print "@input=null, @output=",p.execute(DS_null)
	condition=parseterm("isdatatype @input DS_Array -> forall n:DS_Integer, n<@input -> @output[n]=@input[length @input -n]")
	print condition
	p.randomcheck(condition)


if __name__=="__main__":
	example2b()

