#!/usr/bin/python

import kernel
from kernel import *

tokens=('LPAR','RPAR','ASSIGN','IDENT','WHILE','LCURLY','RCURLY','INTEGER','LBRACKET','RBRACKET','ISNULL','NOT','PLUS','LT','PARSEPROGRAM','PARSETERM','IMPLY','ISDATATYPE','FORALL','COLON','COMMA','EQ','DATATYPE','LENGTH','MINUS','AND')

precedence=(
	#lowest
	('left','PREC_TERM'),
	('right','IMPLY','FORALL','COMMA'),
	('left','AND'),
	('left','NOT'),
	('left','EQ','LT'),
	('left','PLUS','MINUS'),
	('left','LENGTH','ISNULL'),
	#highest
)

t_LPAR		= r'\('
t_RPAR		= r'\)'
t_ASSIGN	= r':='
#t_WHILE		= r'while'
t_LCURLY	= r'{'
t_RCURLY	= r'}'
t_INTEGER	= r'[0-9]+'
t_LBRACKET	= r'[[]'
t_RBRACKET	= r']'
#t_ISNULL	= r'isnull'
t_PLUS		= r'\+'
t_LT		= r'<'
t_IMPLY		= r'->'
t_COLON		= r':'
t_COMMA		= r','
t_EQ		= r'='
t_MINUS		= r'-'

t_PARSEPROGRAM	= r'!!PROGRAM!!'
t_PARSETERM		= r'!!TERM!!'

t_ignore = " \t\n"

keywords={
	"while":"WHILE",
	"isnull":"ISNULL",
	"not":"NOT",
	"isdatatype":"ISDATATYPE",
	"forall":"FORALL",
	"length":"LENGTH",
	"and":"AND",
	}

#t_IDENT		= r'[a-zA-Z0-9_\'@]+'
def t_IDENT(t):
	r'[a-zA-Z@_\'][a-zA-Z0-9_\'@]*'
	kw=keywords.get(t.value)
	if kw is not None:
		t.type=kw
		#print "Got keyword", t

	dt=kernel.__dict__.get(t.value)
	if dt is not None:
		t.type="DATATYPE"
		t.value=dt
	return t



def t_error(t):
	raise Exception( "Illegal character '%s'" % t.value[0] )
	#print "Illegal character '%s'" % t.value[0]
	t.lexer.skip(1)

import ply.lex
lexer=ply.lex.lex()

def p_main(t):
	"""
	main : PARSEPROGRAM program
		| PARSETERM term
	"""
	t[0]=t[2]

def p_program(t):
	"""
	program : statement_list
	"""
	if not isinstance(t[1],list): raise ValueError()
	t[0]=Program(t[1])

"""
term : expression
	| expression IMPLY expression
	| forall variable COLON expression COMMA term
	| DATATYPE
"""
def p_term__expression(t):
	"""
	term : expression %prec PREC_TERM
	"""
	t[0]=t[1]
def p_term__imply(t):
	"""
	term : term IMPLY term
	"""
	t[0]=Product("<None>",t[1],t[3])
def p_term__product(t):
	"""
	term : FORALL variable COLON term COMMA term
	"""
	t[0]=Product(t[2],t[4],t[6])
def p_term__datatype(t):
	"""
	term : DATATYPE
	"""
	t[0]=t[1]


"""
statement_list : 
	| statement statement_list
"""
def p_statement_list__empty(t):
	"""
	statement_list : 
	"""
	t[0]=[]
def p_statement_list__head(t):
	"""
	statement_list : statement statement_list
	"""
	t[0]=[t[1]]+t[2]
	

def p_statement(t):
	"""
	statement : assignment
		| while
	"""
	t[0]=t[1]


def p_statement_block(t):
	"""
	statement_block : statement
		| LCURLY statement_list RCURLY
	"""
	if t[1]==t_LCURLY:
		t[0]=t[2]
	else:
		t[0]=[t[1]]

def p_assignment(t):
	"""
	assignment : assignable ASSIGN expression
	"""
	t[0]=Assignment(t[1],t[3])

def p_while(t):
	"""
	while : WHILE expression statement_block
	"""
	t[0]=While(t[2],t[3])

"""
assignable : variable
	| access
"""
def p_assignable__variable(t):
	"""
	assignable : variable
	"""
	t[0]=Ass_Variable(t[1])
def p_assignable__access(t):
	"""
	assignable : access
	"""
	# access = (variable, position)
	t[0]=Ass_Access(t[1][0],t[1][1])

def p_variable(t):
	"""
	variable : IDENT
	"""
	t[0]=t[1]
	#print "Got variable", t[0]

def p_access(t):
	"""
	access : variable LBRACKET expression RBRACKET
	"""
	t[0]=(t[1],t[3])

"""
expression : variable
	| constant
	| access
	| isnull expression
	| not expression
	| LPAR expression RPAR
	| expression PLUS expression
	| expression MINUS expression
	| expression LT expression
	| ISDATATYPE expression IDENT
	| expression EQ expression
	| LENGTH expression
	| expression AND expression
"""
def p_expression__variable(t):
	"""
	expression : variable
	"""
	t[0]=E_Variable(t[1])
	#print "Got expression",t[0]
def p_expression__constant(t):
	"""
	expression : constant
	"""
	t[0]=E_Constant(t[1])
def p_expression__access(t):
	"""
	expression : access
	"""
	t[0]=E_Access(t[1][0],t[1][1])
def p_expression__isnull(t):
	"""
	expression : ISNULL expression
	"""
	t[0]=E_IsNull(t[2])
def p_expression__not(t):
	"""
	expression : NOT expression
	"""
	t[0]=E_Not(t[2])
def p_expression__PAR(t):
	"""
	expression : LPAR expression RPAR
	"""
	t[0]=t[2]
def p_expression__addition(t):
	"""
	expression : expression PLUS expression
	"""
	t[0]=E_Addition([t[1],t[3]])
def p_expression__substraction(t):
	"""
	expression : expression MINUS expression
	"""
	t[0]=E_Substraction(t[1],t[3])
def p_expression__LT(t):
	"""
	expression : expression LT expression
	"""
	t[0]=E_LT(t[1],t[3])
def p_expression__isdatatype(t):
	"""
	expression : ISDATATYPE expression DATATYPE
	"""
	t[0]=E_IsDataType(t[2],t[3])
def p_expression__equal(t):
	"""
	expression : expression EQ expression
	"""
	t[0]=E_Equal(t[1],t[3])
def p_expression__length(t):
	"""
	expression : LENGTH expression
	"""
	t[0]=E_Length(t[2])
def p_expression__and(t):
	"""
	expression : term AND term
	"""
	t[0]=E_And(t[1],t[3])



"""
constant : INTEGER
	| array
"""
def p_constant__integer(t):
	"""
	constant : INTEGER
	"""
	t[0]=DS_Integer(int(t[1]))
def p_constant__array(t):
	"""
	constant : LBRACKET RBRACKET
	"""
	t[0]=DS_Array()





def p_error(t):
	#print "Syntax error at '%s'" % t.value
	#raise Exception("Syntax error at '%s'" % t.value)
	#print lexer.lexpos
	print "<<<%s>>>!!!<<<%s>>>" % (lexer.lexdata[lexer.lexpos-90:lexer.lexpos],lexer.lexdata[lexer.lexpos:lexer.lexpos+90])
	print "statestack=%s" % (parser.statestack)
	print "symstack=%s" % (parser.symstack)
	#print dir(parser)
	raise Exception("Syntax error at '%s':%s" % (t.value,t.type))

import ply.yacc as yacc
parser=yacc.yacc()

def parseprogram(s):
	#print "parsing",s
	program=parser.parse(t_PARSEPROGRAM+s,lexer=lexer)
	#print "parsed",program
	return program

def parseterm(s):
	#print "parsing",s
	term=parser.parse(t_PARSETERM+s,lexer=lexer)
	#print "parsed",program
	return term

