#!/usr/bin/python

import random
import string
import inspect


class DataStructure:
	def get(self,position):
		return DS_null
	def copy(self):
		#return self
		return self.__class__(self.value)
	@staticmethod
	def randomclass():
		return random.choice(DataStructure.subclasses)
	@classmethod
	def random(cls):
		if cls==DataStructure:
			return cls.randomclass().random()
		else:
			raise NotImplementedError("random() not implemented for class %s" % cls)
class DS_NULL(DataStructure):
	def copy(self):
		return self
	def __str__(self):
		return "NULL"
	@classmethod
	def random(cls):
		return DS_null
DS_null=DS_NULL()
class DS_Bool(DataStructure):
	def __init__(self,value):
		if not isinstance(value,bool): raise ValueError
		self.value=value
	@classmethod
	def random(cls):
		return DS_Bool(random.choice([True,False]))
DS_true=DS_Bool(True)
DS_false=DS_Bool(False)
class DS_Error(DataStructure):
	def __init__(self,info=None):
		self.info=info
	def copy(self):
		return self
	def __str__(self):
		if self.info is None:
			return "<Error>"
		else:
			return "<Error ("+str(self.info)+")>"
	@classmethod
	def random(cls):
		return DS_error
DS_error=DS_Error()
class DS_Integer(DataStructure):
	def __init__(self,value):
		if not isinstance(value,int): raise ValueError
		self.value=value
	def __str__(self):
		return str(self.value)
	@classmethod
	def random(cls):
		return DS_Integer(random.randrange(-1000,1000))#FIXME
class DS_String(DataStructure):
	def __init__(self,value):
		if not isinstance(value,str): raise ValueError
		self.value=value
	def __str__(self):
		return self.value
	@classmethod
	def random(cls):
		#http://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits-in-python
		s=''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(100))#FIXME
		return DS_String(s)
class DS_Array(DataStructure):
	def __init__(self,value=None):
		if value is None:
			self.value=[]
		else:
			self.value=value[:]
	def __str__(self):
		return "["+", ".join(str(x) for x in self.value)+"]"
	def writeat(self,position,value):
		while position>=len(self.value):
			self.value.append(DS_null)
		self.value[position]=value
	def get(self,position):
		if position>=0 and position<len(self.value):
			return self.value[position]
		else:
			return DS_null
	def copy(self):
		return self.__class__(self.value[:])
	@classmethod
	def random(cls):
		n=random.randrange(0,1000)#FIXME
		l=[DS_Integer.random() for t in range(n)]
		return DS_Array(l)
DataStructure.subclasses=[DS_NULL,DS_Bool,DS_Integer,DS_String,DS_Array]

class Assignable: pass #a.k.a. Left value
class Ass_Variable(Assignable):
	def __init__(self,variable):
		self.variable=variable
	def write(self,variables,value):
		variables[self.variable]=value
	def __str__(self):
		return self.variable
class Ass_Access(Assignable):
	def __init__(self,variable,position):
		self.variable=variable #string
		if not isinstance(position,Expression): raise ValueError()
		self.position=position
	def write(self,variables,value):
		pos=self.position.evaluate(variables)
		if not isinstance(pos,DS_Integer):
			raise Exception("%s:%s cannot be used to index arrays" % (pos,pos.__class__))
		#variables[self.variable].value[pos.value]=value
		if self.variable not in variables:
			return DS_Error("Cannot access to position %s of non-existant %s" % (pos.value,self.variable))
		variables[self.variable].writeat(pos.value,value)
	def __str__(self):
		return self.variable+"["+str(self.position)+"]"

class Term: pass

class Expression(Term): pass
class E_Constant(Expression):
	def __init__(self,constant):
		if not isinstance(constant,DataStructure): raise ValueError()
		self.constant=constant
	def evaluate(self,variables):
		return self.constant
	def __str__(self):
		return str(self.constant)
class E_Variable(Expression):
	def __init__(self,variable):
		if not isinstance(variable,str): raise ValueError()
		self.variable=variable
	def evaluate(self,variables):
		if self.variable not in variables:
			raise Exception("There is not variable '%s'" % self.variable)
		return variables[self.variable]
	def __str__(self):
		return self.variable
class E_Access(Expression):
	def __init__(self,variable,position):
		self.variable=variable #string
		if not isinstance(position,Expression): raise ValueError()
		self.position=position
	def evaluate(self,variables):
		if self.variable not in variables:
			raise Exception("There is not variable '%s'" % self.variable)
		pos=self.position.evaluate(variables)
		if isinstance(pos,DS_Error): return pos
		if not isinstance(pos,DS_Integer):
			raise Exception("%s:%s cannot be used to index arrays" % (pos,pos.__class__))
		return variables[self.variable].get(pos.value)
	def __str__(self):
		return self.variable+"["+str(self.position)+"]"
class E_Addition(Expression):
	def __init__(self,le):
		self.le=le
	def evaluate(self,variables):
		#raise Exception("Implement")
		a=0
		for e in self.le:
			data=e.evaluate(variables)
			if not isinstance(data,DS_Integer): return DS_Error("%s cannot be added" % data)
			a+=data.value
		return DS_Integer(a)
	def __str__(self):
		return "+".join( "("+str(e)+")" for e in self.le )
class E_Substraction(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		#raise Exception("Implement")
		data1=self.expression1.evaluate(variables)
		if isinstance(data1,DS_Error): return data1
		data2=self.expression2.evaluate(variables)
		if isinstance(data2,DS_Error): return data2
		if not isinstance(data1,DS_Integer): return DS_Error("%s cannot be added" % data1)
		if not isinstance(data2,DS_Integer): return DS_Error("%s cannot be added" % data2)
		return DS_Integer(data1.value-data2.value)
	def __str__(self):
		return "("+str(self.expression1)+") - ("+str(self.expression2)+")"
class E_IsNull(Expression):
	def __init__(self,expression):
		self.expression=expression
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		if isinstance(data,DS_Error): return data
		return DS_Bool(isinstance(data,DS_NULL))
	def __str__(self):
		return "isnull ("+str(self.expression)+")"
class E_IsDataType(Expression):
	def __init__(self,expression,datatype):
		self.expression=expression
		if datatype.__module__!='kernel': raise Exception("Uh oh!")
		self.datatype=datatype
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		if isinstance(data,DS_Error): return data
		return DS_Bool(data.__class__==self.datatype)
	def __str__(self):
		return "isdatatype (%s) %s" % (self.expression,self.datatype.__name__)
class E_Not(Expression):
	def __init__(self,expression):
		self.expression=expression
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		if isinstance(data,DS_Error): return data
		if not isinstance(data,DS_Bool): raise Exception()
		return DS_Bool(not data.value)
	def __str__(self):
		return "not ("+str(self.expression)+")"
class E_Equal(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		data1=self.expression1.evaluate(variables)
		if isinstance(data1,DS_Error): return data1
		data2=self.expression2.evaluate(variables)
		if isinstance(data2,DS_Error): return data2
		#print "%s(%s)=%s(%s) : %s" % (data1,data1.__class__,data2,data2.__class__,data1==data2)
		if data1.__class__ != data2.__class__: return DS_false
		if data1==data2: return DS_true
		return DS_Bool(data1.value==data2.value)
	def __str__(self):
		s1=str(self.expression1)
		s2=str(self.expression2)
		return "("+s1+")=("+s2+")"
class E_LT(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		data1=self.expression1.evaluate(variables)
		if isinstance(data1,DS_Error): return data1
		data2=self.expression2.evaluate(variables)
		if isinstance(data2,DS_Error): return data2
		if not isinstance(data1,DS_Integer): return DS_null
		if not isinstance(data2,DS_Integer): return DS_null
		#print "%s(%s)=%s(%s) : %s" % (data1,data1.__class__,data2,data2.__class__,data1==data2)
		#if data1.__class__ != data2.__class__: return DS_false
		return DS_Bool(data1.value<data2.value)
	def __str__(self):
		s1=str(self.expression1)
		s2=str(self.expression2)
		return "("+s1+")<("+s2+")"
class E_LE(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		data1=self.expression1.evaluate(variables)
		if isinstance(data1,DS_Error): return data1
		data2=self.expression2.evaluate(variables)
		if isinstance(data2,DS_Error): return data2
		if not isinstance(data1,DS_Integer): return DS_null
		if not isinstance(data2,DS_Integer): return DS_null
		#print "%s(%s)=%s(%s) : %s" % (data1,data1.__class__,data2,data2.__class__,data1==data2)
		#if data1.__class__ != data2.__class__: return DS_false
		return DS_Bool(data1.value<=data2.value)
	def __str__(self):
		s1=str(self.expression1)
		s2=str(self.expression2)
		return "("+s1+")<=("+s2+")"
class E_Length(Expression):
	def __init__(self,expression):
		self.expression=expression
	def evaluate(self,variables):
		data=self.expression.evaluate(variables)
		if isinstance(data,DS_Error): return data
		if not isinstance(data,DS_Array): return DS_Error("Calculating length of the non-vector %s" % data)
		return DS_Integer(len(data.value))
	def __str__(self):
		return "length ("+str(self.expression)+")"
class E_And(Expression):
	def __init__(self,expression1,expression2):
		self.expression1=expression1
		self.expression2=expression2
	def evaluate(self,variables):
		data1=self.expression1.evaluate(variables)
		if isinstance(data1,DS_Error): return data1
		data2=self.expression2.evaluate(variables)
		if isinstance(data2,DS_Error): return data2
		if not isinstance(data1,DS_Bool): raise Exception()
		if not isinstance(data2,DS_Bool): raise Exception()
		return DS_Bool(data1.value and data2.value)
	def __str__(self):
		return "("+str(self.expression1)+") and ("+str(self.expression2)+")"

class BoundTerm(Term):
	def __init__(self,var,vartype,body):
		if self.__class__==BoundTerm:
			raise ValueError("You cannot instantiate a BoundTerm")
		self.variable=var
		#XXX variable = None for independent product
		self.vartype=vartype
		self.body=body
		#bfv=set(body.getfreevars())
		#bfv.discard(var.name)
		#self.freevars=bfv|vartype.getfreevars()

class Product(BoundTerm):
	"""
	forall var:vartype, body
	"""
	def evaluate(self,variables):
		head=self.vartype
		headdata=None
		if isinstance(head,Expression):
			headdata=self.vartype.evaluate(variables)
		#print "vartype=%s" % data
		if isinstance(headdata,DS_Bool):
			if not headdata.value: return DS_true #False -> anything
			return self.body.evaluate(variables)
		elif isinstance(headdata,DS_NULL):
			d=dict(variables)
			variables[self.variable]=DS_null
			r=self.body.evaluate(variables)
			if not isinstance(r,DS_Bool): raise Exception()
			return r
		elif head==DS_Integer:
			#FIXME: induction
			for i in range(-100,100):
				d=dict(variables)
				#print d["@input"].value[0],d["@output"].value[0]
				variables[self.variable]=DS_Integer(i)
				#print self.body
				r=self.body.evaluate(variables)
				if not isinstance(r,DS_Bool): raise Exception()
				if not r.value: return DS_false
			return DS_true
		else:
			raise Exception("Cannot evaluate such product (%s:%s)" % (headdata,headdata.__class__))
	def __str__(self):
		if inspect.isclass(self.vartype):
			if self.vartype.__module__!='kernel': raise Exception("Uh oh!")
			str_vartype=self.vartype.__name__
		else:
			str_vartype=str(self.vartype)
		if self.variable is None:
			return "%s -> %s" % (str_vartype,self.body)
		else:
			return "forall %s:%s, %s" % (self.variable,str_vartype,self.body)

class Abstraction(BoundTerm): pass
class Application(Term): pass

class Statement: pass
class Assignment(Statement):
	def __init__(self,ass,expression):
		# ass := expression
		if not isinstance(ass,Assignable): raise ValueError("ass=%s:%s should be an Assignable" % (ass,ass.__class__))
		if not isinstance(expression,Expression): raise ValueError("expression=%s:%s should be an Expression" % (expression,expression.__class__))
		self.ass=ass
		self.expression=expression
	def execute(self,variables):
		#variables[self.variable]=self.expression.evaluate(variables)
		return self.ass.write(variables,self.expression.evaluate(variables).copy())
	def __str__(self):
		return "%s := %s" % (str(self.ass),str(self.expression))
class While(Statement):
	def __init__(self,condition,body):
		if not isinstance(condition,Expression): raise ValueError()
		self.condition=condition
		self.body=body #list of Statements
	def execute(self,variables):
		while True:
			e=self.condition.evaluate(variables)
			if not isinstance(e,DS_Bool): return DS_Error("Expression (%s) with value (%s) in while loop is not a DS_Bool" % (self.condition,e))
			if not e.value: return
			for s in self.body:
				s.execute(variables)
	def __str__(self):
		r="while "+str(self.condition)+"\n{\n"
		for s in self.body:
			r+="\t"+str(s)+"\n"
		r+="}"
		return r


class Program:
	"""
	Class representing a program, which receives some @input and returns some @output
	"""
	def __init__(self,body):
		self.body=body #list of Statement
	def execute(self,input):
		"""
		Executes the program over the input and returns the output
		"""
		variables={"@input":input}
		for s in self.body:
			r=s.execute(variables)
			if isinstance(r,DS_Error): return r
		if "@output" in variables:
			return variables["@output"]
		else:
			return DS_Error("No output generated")
	def randomcheck(self,condition,n=1000):
		"""
		Executes the program for n random inputs and counts how many times condition holds
		"""
		good=0
		badshown=0
		for i in range(n):
			input=DataStructure.random()
			#print "INPUT:",input
			output=self.execute(input)
			variables={"@input":input,"@output":output}
			#print "OUTPUT:",output,id(output)
			e=condition.evaluate(variables)
			if isinstance(e,DS_Bool) and e.value:
				good+=1
				#print "isgood"
			else:
				pass
				if badshown<3:
					print
					print "BAD INPUT:", input
					print "BAD OUTPUT:", output
					badshown+=1
		print "good = %d/%d (%f)" % (good,n,float(good)/n)
	def __str__(self):
		r=""
		for s in self.body:
			r+=str(s)+"\n"
		return r



def programforgoal(goal,hypotheses):
	#print "--------"
	#print goal
	#print "["+"; ".join((str(x) for x in hypotheses))+"]"
	if isinstance(goal,Product):
		#print goal.variable
		#print goal.vartype
		#print goal.body
		if isinstance(goal.vartype,Term):
			H=hypotheses+[goal.vartype]
			return programforgoal(goal.body,H)
		elif goal.vartype==DS_Integer:
			rec=programforgoal(goal.body,hypotheses)
			#goal.body should be true except for some region
			min=E_Constant(DS_Integer(0))
			max=E_Length(E_Variable("@input"))
			v_0=Assignment(Ass_Variable(goal.variable),min)
			v_inc=Assignment(Ass_Variable(goal.variable),E_Addition([E_Variable(goal.variable),E_Constant(DS_Integer(1))]))
			w=While(E_LT(E_Variable(goal.variable),max),rec.body+[v_inc])
			return Program([v_0,w])
		else:
			raise Exception("Implement vartype=%s" % str(goal.vartype))
	elif isinstance(goal,E_Equal):
		#maybe a simple assignment?
		#print " %s   =   %s" % (goal.expression1,goal.expression2)
		if isinstance(goal.expression1,E_Variable) and goal.expression1.variable!="@input":
			v=Ass_Variable(goal.expression1.variable)
			a=Assignment(v,goal.expression2)
			return Program([a])
		if isinstance(goal.expression1,E_Access) and goal.expression1.variable!="@input":
			v=Ass_Access(goal.expression1.variable,goal.expression1.position)
			a=Assignment(v,goal.expression2)
			return Program([a])
		raise Exception("Could not transform equality into assignment")
	raise Exception("goal=%s:%s" % (goal,goal.__class__))


def generateprogram(condition):
	"""
	Search for a program satisfying given condition
	"""
	hypotheses=[]
	term=condition
	while isinstance(term,Product) and term.variable is None:
		hypotheses.append(term.vartype)
		print "Found hypothesis", term.vartype
		term=term.body
	print "Looking for", term
	p=programforgoal(term,hypotheses)
	o_init=Assignment(Ass_Variable("@output"),E_Constant(DS_Array()))
	p=Program([o_init]+p.body)
	return p






